/*
  moors.cpp - Make use of the moors for player ships.

  Jason Hood, 9 February, 2010.

  Requires and assumes v1.1.
*/

#include "Server.h"
#include <vector>

#define NAKED	__declspec(naked)
#define STDCALL __stdcall


PBYTE server, content;
DWORD old_content, old_launch_pos;

const UINT can_use_berths = 0x87309f87; // CreateID( "can_use_berths" );


// Turn the ship around on launch, to face the moor.
#define ADDR_LAUNCH_POS ((PDWORD)(0x62aa9de+1))

// Recognise the moors as valid docking points.
#define ADDR_HAS_DOCK	((PBYTE)0x62c08a8+2)

// Use mooring fixture for launching.
#define ADDR_FIX_LAUNCH (server + 0x6d00a6b - 0x6ce0000)

// Dock at the moor.
#define ADDR_DOCK	(server + 0x6d0c24c - 0x6ce0000)

// Determine the address of Content.dll.
#define ADDR_CONTENT	((PDWORD)(server + 0x6d44c2d+1 - 0x6ce0000))

// Use moors for launching (from a station; only needed for SP).
#define ADDR_STN_LAUNCH ((PUSHORT)(content + 0x6f5e2a0 - 0x6ea0000))


DWORD dummy;
#define ProtectX( addr, size ) \
  VirtualProtect( addr, size, PAGE_EXECUTE_READWRITE, &dummy );

#define RELOFS( from, to ) \
  *(PDWORD)(from) = (DWORD)(to) - (DWORD)(from) - 4;

#define CALL( from, to ) \
  *(PBYTE)(from) = 0xe8; \
  RELOFS( (DWORD)from+1, to )


void Patch_Content()
{
  content = (PBYTE)GetModuleHandle( "content.dll" );
  ProtectX( ADDR_STN_LAUNCH, 2 );
  *ADDR_STN_LAUNCH = 0xC030; // xor al, al ; always use multiplayer code
}


NAKED
void Content_Hook()
{
  __asm {
	push	edi
	call	old_content
	add	esp, 4
	call	Patch_Content
	ret
  }
}


NAKED
void Launch_Pos_Hook()
{
  __asm {
	push	ebx
	push	[esp+12+4]
	push	edx
	push	eax
	call	old_launch_pos
	add	esp, 16

	mov	edx, [edi+0x88]
	mov	eax, [esp+0x88+0x10+4]
	shl	eax, 4
	mov	eax, [edx+eax+12]	// this is different to the other type
	cmp	eax, 4
	je	moor
	cmp	eax, 5
	jne	done
  moor:
	// It's a moor, so modify the orientation to turn around.
	mov	al, 0x80
	xor	[ebx+ 0+3], al
	xor	[ebx+ 4+3], al
	xor	[ebx+ 8+3], al
	xor	[ebx+24+3], al
	xor	[ebx+28+3], al
	xor	[ebx+32+3], al
  done:
	ret
  }
}


NAKED
void Dock_Hook()
{
  __asm {
	mov	al, [ebx+0x1c]
	test	al, al
	jnz	done
	// It's false, so a safe bet that it's a moor.  Is it the player?
	mov	eax, [edi]
	mov	ecx, edi
	call	[eax+0xbc] // is_player
	test	al, al
  done:
	ret
  }
}


typedef std::vector<UINT> UIntVec;
typedef bool (*TGetMissionProperty)( UINT, UIntVec& );

int STDCALL Use_Berths( UINT player )
{
  UINT ship_id;
  player /= 0x418;
  pub::Player::GetShipID( player, ship_id );
  UIntVec prop;
  TGetMissionProperty
   GetMissionProperty = (TGetMissionProperty)(content + 0x6f587d0 - 0x6ea0000);
  GetMissionProperty( ship_id, prop );
  for (int i = prop.size(); --i >= 0;)
    if (prop[i] == can_use_berths)
      return 0;

  return 4;
}


NAKED
void Fix_Launch_Hook()
{
  __asm {
	cmp	eax, 1
	jg	moor
	mov	edx, [esp+0x20+4]
	mov	eax, [edx]
	ret
  moor:
	push	esi
	call	Use_Berths
	mov	edx, [esp+0x20+4]
	mov	eax, [edx+eax]
	ret
  }
}


void Patch()
{
  server = (PBYTE)GetModuleHandle( "server.dll" );

  ProtectX( ADDR_LAUNCH_POS, 4 );
  ProtectX( ADDR_HAS_DOCK,   7 );

  ProtectX( ADDR_FIX_LAUNCH, 6 );
  ProtectX( ADDR_DOCK,	     5 );
  ProtectX( ADDR_CONTENT,    4 );

  // Hook in code to get the address of content.dll.
  old_content = *ADDR_CONTENT + (DWORD)ADDR_CONTENT + 4;
  RELOFS( ADDR_CONTENT, Content_Hook );

  // Add the moors (types 4, 5 & 6) to the dock points.
  memcpy( ADDR_HAS_DOCK, "\x02\x7e\x05\x83\xf9\x07\x7e", 7 );

  // Launching from a moor needs to turn the ship around.
  old_launch_pos = *ADDR_LAUNCH_POS + (DWORD)ADDR_LAUNCH_POS + 4;
  RELOFS( ADDR_LAUNCH_POS, Launch_Pos_Hook );

  // Actually dock at the moor.
  CALL( ADDR_DOCK, Dock_Hook );

  // Launch from the moor.
  CALL( ADDR_FIX_LAUNCH, Fix_Launch_Hook );
  ADDR_FIX_LAUNCH[5] = 0x90;
}


BOOL WINAPI DllMain( HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved )
{
  if (fdwReason == DLL_PROCESS_ATTACH)
    Patch();

  return TRUE;
}
